header "fifo" /{

typedef struct {
	int opened;
	char *base;
	int bufsize;
	int start;
        int sleeps;
	int len;
	struct wait_queue *wait;
	struct tq_struct wake_up_task;	

	struct semaphore *semid;
} ddk_fifo_t;

#define DDK_FIFO_NO 4

#define DDK_ADDR(minor)         (&ddk_fifoes[minor])

#define DDK_OPEN(minor)         (DDK_ADDR(minor)->open)
#define DDK_BASE(minor)         (DDK_ADDR(minor)->base)
#define DDK_BUF(minor)          (DDK_ADDR(minor)->bufsize)
#define DDK_START(minor)        (DDK_ADDR(minor)->start)
#define DDK_LEN(minor)          (DDK_ADDR(minor)->len)
#define DDK_FREE(minor)         (DDK_BUF(minor) - DDK_LEN(minor))
#define DDK_SLEEPS(minor)	(DDK_ADDR(minor)->sleeps)
#define DDK_SEMID(minor)	(DDK_ADDR(minor)->semid)

#define DDK_WAKE_UP_TASK(minor)	(DDK_ADDR(minor)->wake_up_task)
#define DDK_WRAP(minor,pos)     ((pos) < DDK_BUF(minor)? (pos) : (pos) - DDK_BUF(minor))
#define DDK_END(minor)          DDK_WRAP(minor, DDK_START(minor)+DDK_LEN(minor))

#define DDK_EMPTY(minor)        (DDK_LEN(minor)==0)
#define DDK_FULL(minor)         (DDK_FREE(minor)==0)
#define DDK_MAX_RCHUNK(minor)   (DDK_BUF(minor) - DDK_START(minor))
#define DDK_MAX_WCHUNK(minor)   (DDK_BUF(minor) - DDK_END(minor))


}/

file "fifo" /{

static ddk_fifo_t ddk_fifoes[DDK_NR_FIFO];


}/


func FifoResize (unsigned int minor, int size ) /{
	void *mem;
	void *old;

        if (minor >= DDK_FIFO_NO ) {
                return -ENODEV;
        }
        mem = kmalloc(size, GFP_KERNEL);
        if (!mem) {
                return -ENOMEM;
        }
        old = DDK_BASE(minor);

        DDK_BASE(minor) = mem;
        DDK_BUF(minor) = size;
        DDK_START(minor) = 0;
        DDK_LEN(minor) = 0;
	DDK_SEMID(minor)->count = 1;

        if (DDK_OPEN(minor)) {
                kfree(old);
        }
        return size;
}/

func FifoCreate (unsigned int minor, int size ) /{
        int result;

        if (minor >= DDK_FIFO_NO) {
                return -ENODEV;
        }
        if (DDK_OPEN(minor)) {
                return -EBUSY;
        }
        if ((result = %<driver.modname>_FifoResize(minor, size)) < 0) {
                return result;
        }
        DDK_OPEN(minor) = 1;
        DDK_SLEEPS(minor) = 0;

        return 0;
}/

func FifoDestroy (unsigned int minor, int size ) /{

        if (minor >= DDK_FIFO_NO) {
                return -ENODEV;
        }
        if (!DDK_OPEN(minor)) {
                return -EINVAL;
        }
        DDK_OPEN(minor) = 0;
        kfree(DDK_BASE(minor));
        return 0;

}/

func FifoPut( unsigned int minor, void *buf, int count) /{
        int chars = 0, free = 0, written = 0;
        char *pipebuf;

        if (minor >= DDK_FIFO_NO) {
                return -ENODEV;
        }
        if (!DDK_OPEN(minor))
                return -EINVAL;

        if (DDK_FREE(minor) < count) {
                r_restore_flags(flags);
                return -ENOSPC;
        }
        while (count > 0 && (free = DDK_FREE(minor))) {
                chars = DDK_MAX_WCHUNK(minor);
                if (chars > count)
                        chars = count;
                if (chars > free)
                        chars = free;
                pipebuf = DDK_BASE(minor) + DDK_END(minor);
                written += chars;
                DDK_LEN(minor) += chars;
                count -= chars;
                memcpy(pipebuf, buf, chars);
                buf += chars;
        }
#if 0
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
#endif


        return count;

}/

func FifoGet( unsigned int minor, void *buf, int count) /{
        long flags;
        int chars = 0, size = 0, read = 0;
        char *pipebuf;

        if (minor >= DDK_FIFO_NO) {
                return -ENODEV;
        }
        if (!DDK_OPEN(minor))
                return -EINVAL;

        while (count > 0 && (size = DDK_LEN(minor))) {
                chars = DDK_MAX_RCHUNK(minor);
                if (chars > count)
                        chars = count;
                if (chars > size)
                        chars = size;

                read += chars;
                pipebuf = DDK_BASE(minor) + DDK_START(minor);
                DDK_START(minor) += chars;
                DDK_START(minor) = DDK_WRAP(minor, DDK_START(minor));
                DDK_LEN(minor) -= chars;
                count -= chars;
                memcpy(buf, pipebuf, chars);
                buf += chars;
        }
#if 0
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
#endif
        return read;

}/


func FifoRead(struct inode *inode, struct file *filp, char *buf, int count)
/{
	unsigned int minor = MINOR(inode->i_rdev);
	long flags;
	int result;

	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (filp->f_flags & O_NONBLOCK) {
		/*      if (DDK_LOCK(minor))
		   return -EAGAIN;
		 */
		if (DDK_EMPTY(minor))
			return 0;
	} else
		while (DDK_EMPTY(minor) /* || DDK_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return -ERESTARTSYS;
			set_bit(1, &DDK_SLEEPS(minor));
			interruptible_sleep_on(&DDK_WAIT(minor));
		}
/*      DDK_LOCK(minor)++; */
	while (count > 0 && (size = DDK_LEN(minor))) {
		chars = DDK_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = DDK_BASE(minor) + DDK_START(minor);
		count -= chars;
		memcpy_tofs(buf, pipebuf, chars);
		r_save_flags(flags);
		r_cli();
		DDK_START(minor) += chars;
		DDK_START(minor) = DDK_WRAP(minor, DDK_START(minor));
		DDK_LEN(minor) -= chars;
		r_restore_flags(flags);
		buf += chars;
	}
/*      DDK_LOCK(minor)--; */
	if (read) {
		inode->i_atime = CURRENT_TIME;
		if (DDK_HANDLER(minor)) {
			if ((result = (*DDK_HANDLER(minor)) (minor)) < 0) {
				return result;
			}
		}
		return read;
	}
	return 0;
}/


func FifoWrite(struct inode *inode, struct file *filp, const char *buf, int count)
/{
	unsigned int minor = MINOR(inode->i_rdev);
	long flags;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;
	int result;

	if (count <= DDK_BUF(minor))
		free = count;
	else
		free = 1;

	while (count > 0) {
		while ((DDK_FREE(minor) < free) /* || DDK_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return written ? : -ERESTARTSYS;
			if (filp->f_flags & O_NONBLOCK)
				return written ? : -EAGAIN;
			set_bit(1, &DDK_SLEEPS(minor));
			interruptible_sleep_on(&DDK_WAIT(minor));
		}
		/*      DDK_LOCK(minor)++; */
		while (count > 0 && (free = DDK_FREE(minor))) {
			chars = DDK_MAX_WCHUNK(minor);
			if (chars > count)
				chars = count;
			if (chars > free)
				chars = free;
			r_save_flags(flags);
			r_cli();
			pipebuf = DDK_BASE(minor) + DDK_END(minor);
			r_restore_flags(flags);
			count -= chars;
			written += chars;

			memcpy_fromfs(pipebuf, buf, chars);

			r_save_flags(flags);
			r_cli();
			DDK_LEN(minor) += chars;
			r_restore_flags(flags);
			buf += chars;
		}
		/*      DDK_LOCK(minor)--; */
		free = 1;
	}
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	if (DDK_HANDLER(minor)) {
		if ((result = (*DDK_HANDLER(minor)) (minor)) < 0) {
			return result;
		}
	}
	return written;
}/

func FifoSelect(struct inode *inode, struct file *filp, int sel_type, select_table * wait)
/{
	unsigned int minor = MINOR(inode->i_rdev);
	switch (sel_type) {
		case SEL_IN:
			if (!DDK_EMPTY(minor))
				return 1;
			select_wait(&DDK_WAIT(minor), wait);
			return 0;
		case SEL_OUT:
			if (DDK_EMPTY(minor))
				return 1;
			select_wait(&DDK_WAIT(minor), wait);
			return 0;
	}
	return 0;
}/

